/*  A weekend project to try to emulate the original RAND text editor
    Copyright (C) 2013  Mike Stabenfeldt

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <alloca.h>

#include "edit.h"

extern globalStruct globalData;


runCmd(int argc, char **argv)
{
char tmpName1[50], tmpName2[50], *runCmd, *line;
long long startLineNo;
int c, i, linesOut, linesIn, maxLen, len;
lineStruct *start, *end=NULL, *ls, *next, *last, *beforeStart, *afterEnd;
windowStruct *w=globalData.activeWindow;
fileStruct *fs;
FILE *fp;

    if( !(fs=w->fileInfo) )
	{
	emsg("Open a file before replace");
	return(0);
	}
    if( argc < 2 )
	{
	emsg("run: program [args]");
	return(0);
	}
    /* Create a tmp file with the input data */
    /* Note an area must be marked or just the current line is assumed */
    /* Marked area must be a set of lines, not a rect */
    getCursorLine(w, &ls);
    start = end = ls;
    linesOut = 0;
    startLineNo = w->topOffset + w->cursorLineNo;
    if( globalData.markSet )
	{
	if( w->leftOffset + w->cursorX != globalData.markRefX )
	    {
	    dingMsg("Can't \"run\" in rect region");
	    updateCursor(w);
	    return(0);
	    }
	if( w->topOffset + w->cursorLineNo < globalData.markRefY )
	    {
	    getLine(w, fs, globalData.markRefY-1, &end);
	    }
	else
	    {
	    startLineNo = globalData.markRefY;
	    getLine(w, fs, globalData.markRefY-1, &ls);
	    start = ls;
	    }
	globalData.markSet = 0;
	}

    if( !canModifyFile(fs) )
	{
	updateCursor(w);
	return(0);
	}
    sprintf(tmpName1, "/tmp/sre_out%d", getpid());
    sprintf(tmpName2, "/tmp/sre_in%d", getpid());
    if( !(fp=fopen(tmpName1, "w")) )
	{
	dingMsg("Can't create tmp file");
	updateCursor(w);
	return(0);
	}
    for( ; ls ; ls=ls->next)
	{
	unpackDiskLine(fs, ls);
	fprintf(fp, "%s\n", ls->data.line);
	linesOut++;
	if( end == ls )
	    break;
	}
    fclose(fp);

    len = strlen(tmpName1) + strlen(tmpName2) + 10;
    for(i=1; i<argc; i++)
	len += strlen(argv[i])+2;
    runCmd = (char *) alloca(len);
    len=0;
    for(i=1; i<argc; i++)
	len+= sprintf( &runCmd[len], "%s ", argv[i]);
    len+= sprintf(&runCmd[len], " < %s", tmpName1);
    len+= sprintf(&runCmd[len], " > %s", tmpName2);
    dingMsg(runCmd);

    /* Cheat, use the system command */
    i=system(runCmd);
    unlink(tmpName1);
    if( i != 0 )
	{
	sprintf(runCmd, "%s failed", argv[1]);
	dingMsg(runCmd);
	redrawWindow(w);
	return(0);
	}
    if( (fp=fopen(tmpName2, "r")) == NULL )
	{
	sprintf(runCmd, "%s null output", argv[1]);
	dingMsg(runCmd);
	redrawWindow(w);
	return(0);
	}
    maxLen = len = 0;
    for(linesIn=0; (c=getc(fp)) != EOF; )
	{
	if( c == '\n' )
	    {
	    if( len > maxLen )
		maxLen = len;
	    linesIn++;
	    }
	else
	    len++;
	}
    beforeStart = start->prev;
    afterEnd = end->next;
    /* Remove the old data */
    ls = start;
    fs->numLines -= linesOut;

    if( w->topLine == start )
	w->topLine = afterEnd;
    if( fs->head == start )
	{
	afterEnd->prev = NULL;
	fs->head = afterEnd;
	}
    else
	beforeStart->next = afterEnd;
    if( afterEnd )
	afterEnd->prev = beforeStart;
    end->next = NULL;

    /* Now delete the lines */
    convertToLineBuffer(w->fileInfo, start, linesOut, &(globalData.runBuffer), 1);

    /* Now load up the new data */
    line = (char *) alloca(maxLen+2);
    rewind(fp);
    last = beforeStart;
    while(fgets(line, maxLen+1, fp) != NULL)
	{
	ls = (lineStruct *) calloc(sizeof(lineStruct), 1);
	len = strlen(line);
	if( line[len-1] == '\n')
	    {
	    line[len-1] = '\0';
	    len--;
	    }
	if( len < 80 )
	    {
	    ls->data.line = (char *) malloc(80);
	    ls->numCharsAvail = 80;
	    }
	else
	    {
	    ls->data.line = (char *) malloc(len);
	    ls->numCharsAvail = len;
	    }
	ls->numChars = len;
	strcpy(ls->data.line, line);
	if( last )
	    last->next = ls;
	else
	    fs->head = ls;
	ls->prev = last;
	last = ls;
	fs->numLines++;
	}
    ls->next = afterEnd;
    if( afterEnd )
	afterEnd->prev = ls;
    else
	fs->tail = ls;
    sprintf(runCmd, "%s in %d out %d", argv[1], linesIn, linesOut);
    dingMsg(runCmd);
    w->cursorLine = NULL;
    gotoline(w, startLineNo);
    redrawWindow(w);
    return(1);
}
